"""
Some codes from https://github.com/Newmu/dcgan_code
"""
from __future__ import division
import math
import json
import random
import pprint
import scipy.misc
import numpy as np
from time import gmtime, strftime
from six.moves import xrange
from glob import glob
import os

from ops import *
from utils import *

import tensorflow as tf
import tensorflow.contrib.slim as slim

pp = pprint.PrettyPrinter()

get_stddev = lambda x, k_h, k_w: 1/math.sqrt(k_w*k_h*x.get_shape()[-1])

def show_all_variables():
  model_vars = tf.trainable_variables()
  slim.model_analyzer.analyze_vars(model_vars, print_info=True)

def get_image(image_path, input_height, input_width,
              resize_height=64, resize_width=64,
              crop=True, grayscale=False):
  image = imread(image_path, grayscale)
  return transform(image, input_height, input_width,
                   resize_height, resize_width, crop)

def save_images(images, size, image_path):
  return imsave(inverse_transform(images), size, image_path)

def imread(path, grayscale = False):
  if (grayscale):
    return scipy.misc.imread(path, flatten = True).astype(np.float)
  else:
    return scipy.misc.imread(path).astype(np.float)

def merge_images(images, size):
  return inverse_transform(images)

def merge(images, size):
  h, w = images.shape[1], images.shape[2]
  if (images.shape[3] in (3,4)):
    c = images.shape[3]
    img = np.zeros((h * size[0], w * size[1], c))
    for idx, image in enumerate(images):
      i = idx % size[1]
      j = idx // size[1]
      img[j * h:j * h + h, i * w:i * w + w, :] = image
    return img
  elif images.shape[3]==1:
    img = np.zeros((h * size[0], w * size[1]))
    for idx, image in enumerate(images):
      i = idx % size[1]
      j = idx // size[1]
      img[j * h:j * h + h, i * w:i * w + w] = image[:,:,0]
    return img
  else:
    raise ValueError('in merge(images,size) images parameter '
                     'must have dimensions: HxW or HxWx3 or HxWx4')

def imsave(images, size, path):
  image = np.squeeze(merge(images, size))
  return scipy.misc.imsave(path, image)

def center_crop(x, crop_h, crop_w,
                resize_h=64, resize_w=64):
  if crop_w is None:
    crop_w = crop_h
  h, w = x.shape[:2]
  j = int(round((h - crop_h)/2.))
  i = int(round((w - crop_w)/2.))
  return scipy.misc.imresize(
      x[j:j+crop_h, i:i+crop_w], [resize_h, resize_w])

def transform(image, input_height, input_width, 
              resize_height=64, resize_width=64, crop=True):
  if crop:
    cropped_image = center_crop(
      image, input_height, input_width, 
      resize_height, resize_width)
  else:
    cropped_image = scipy.misc.imresize(image, [resize_height, resize_width])
  return np.array(cropped_image)/127.5 - 1.

def inverse_transform(images):
  return (images+1.)/2.

def to_json(output_path, *layers):
  with open(output_path, "w") as layer_f:
    lines = ""
    for w, b, bn in layers:
      layer_idx = w.name.split('/')[0].split('h')[1]

      B = b.eval()

      if "lin/" in w.name:
        W = w.eval()
        depth = W.shape[1]
      else:
        W = np.rollaxis(w.eval(), 2, 0)
        depth = W.shape[0]

      biases = {"sy": 1, "sx": 1, "depth": depth, "w": ['%.2f' % elem for elem in list(B)]}
      if bn != None:
        gamma = bn.gamma.eval()
        beta = bn.beta.eval()

        gamma = {"sy": 1, "sx": 1, "depth": depth, "w": ['%.2f' % elem for elem in list(gamma)]}
        beta = {"sy": 1, "sx": 1, "depth": depth, "w": ['%.2f' % elem for elem in list(beta)]}
      else:
        gamma = {"sy": 1, "sx": 1, "depth": 0, "w": []}
        beta = {"sy": 1, "sx": 1, "depth": 0, "w": []}

      if "lin/" in w.name:
        fs = []
        for w in W.T:
          fs.append({"sy": 1, "sx": 1, "depth": W.shape[0], "w": ['%.2f' % elem for elem in list(w)]})

        lines += """
          var layer_%s = {
            "layer_type": "fc", 
            "sy": 1, "sx": 1, 
            "out_sx": 1, "out_sy": 1,
            "stride": 1, "pad": 0,
            "out_depth": %s, "in_depth": %s,
            "biases": %s,
            "gamma": %s,
            "beta": %s,
            "filters": %s
          };""" % (layer_idx.split('_')[0], W.shape[1], W.shape[0], biases, gamma, beta, fs)
      else:
        fs = []
        for w_ in W:
          fs.append({"sy": 5, "sx": 5, "depth": W.shape[3], "w": ['%.2f' % elem for elem in list(w_.flatten())]})

        lines += """
          var layer_%s = {
            "layer_type": "deconv", 
            "sy": 5, "sx": 5,
            "out_sx": %s, "out_sy": %s,
            "stride": 2, "pad": 1,
            "out_depth": %s, "in_depth": %s,
            "biases": %s,
            "gamma": %s,
            "beta": %s,
            "filters": %s
          };""" % (layer_idx, 2**(int(layer_idx)+2), 2**(int(layer_idx)+2),
               W.shape[0], W.shape[3], biases, gamma, beta, fs)
    layer_f.write(" ".join(lines.replace("'","").split()))

def make_gif(images, fname, duration=2, true_image=False):
  import moviepy.editor as mpy

  def make_frame(t):
    try:
      x = images[int(len(images)/duration*t)]
    except:
      x = images[-1]

    if true_image:
      return x.astype(np.uint8)
    else:
      return ((x+1)/2*255).astype(np.uint8)

  clip = mpy.VideoClip(make_frame, duration=duration)
  clip.write_gif(fname, fps = len(images) / duration)

def visualize(sess, dcgan, config, option):
  image_frame_dim = int(math.ceil(config.batch_size**.5))
  if option == 0:
    z_sample = np.random.uniform(-0.5, 0.5, size=(config.batch_size, dcgan.z_dim))
    samples = sess.run(dcgan.sampler, feed_dict={dcgan.z: z_sample})
    save_images(samples, [image_frame_dim, image_frame_dim], './samples/test_%s.png' % strftime("%Y%m%d%H%M%S", gmtime()))
  elif option == 1:
    values = np.arange(0, 1, 1./config.batch_size)
    for idx in xrange(100):
      print(" [*] %d" % idx)
      z_sample = np.zeros([config.batch_size, dcgan.z_dim])
      for kdx, z in enumerate(z_sample):
        z[idx] = values[kdx]

      if config.dataset == "mnist":
        y = np.random.choice(10, config.batch_size)
        y_one_hot = np.zeros((config.batch_size, 10))
        y_one_hot[np.arange(config.batch_size), y] = 1

        samples = sess.run(dcgan.sampler, feed_dict={dcgan.z: z_sample, dcgan.y: y_one_hot})
      else:
        samples = sess.run(dcgan.sampler, feed_dict={dcgan.z: z_sample})

      save_images(samples, [image_frame_dim, image_frame_dim], './samples/test_arange_%s.png' % (idx))
  elif option == 2:
    values = np.arange(0, 1, 1./config.batch_size)
    for idx in [random.randint(0, 99) for _ in xrange(100)]:
      print(" [*] %d" % idx)
      z = np.random.uniform(-0.2, 0.2, size=(dcgan.z_dim))
      z_sample = np.tile(z, (config.batch_size, 1))
      #z_sample = np.zeros([config.batch_size, dcgan.z_dim])
      for kdx, z in enumerate(z_sample):
        z[idx] = values[kdx]

      if config.dataset == "mnist":
        y = np.random.choice(10, config.batch_size)
        y_one_hot = np.zeros((config.batch_size, 10))
        y_one_hot[np.arange(config.batch_size), y] = 1

        samples = sess.run(dcgan.sampler, feed_dict={dcgan.z: z_sample, dcgan.y: y_one_hot})
      else:
        samples = sess.run(dcgan.sampler, feed_dict={dcgan.z: z_sample})

      try:
        make_gif(samples, './samples/test_gif_%s.gif' % (idx))
      except:
        save_images(samples, [image_frame_dim, image_frame_dim], './samples/test_%s.png' % strftime("%Y%m%d%H%M%S", gmtime()))
  elif option == 3:
    values = np.arange(0, 1, 1./config.batch_size)
    for idx in xrange(100):
      print(" [*] %d" % idx)
      z_sample = np.zeros([config.batch_size, dcgan.z_dim])
      for kdx, z in enumerate(z_sample):
        z[idx] = values[kdx]

      samples = sess.run(dcgan.sampler, feed_dict={dcgan.z: z_sample})
      make_gif(samples, './samples/test_gif_%s.gif' % (idx))
  elif option == 4:
    image_set = []
    values = np.arange(0, 1, 1./config.batch_size)

    for idx in xrange(100):
      print(" [*] %d" % idx)
      z_sample = np.zeros([config.batch_size, dcgan.z_dim])
      for kdx, z in enumerate(z_sample): z[idx] = values[kdx]

      image_set.append(sess.run(dcgan.sampler, feed_dict={dcgan.z: z_sample}))
      make_gif(image_set[-1], './samples/test_gif_%s.gif' % (idx))

    new_image_set = [merge(np.array([images[idx] for images in image_set]), [10, 10]) \
        for idx in range(64) + range(63, -1, -1)]
    make_gif(new_image_set, './samples/test_gif_merged.gif', duration=8)

  elif option == 5:
    batch_num = 1
    input_w = 500
    input_h = 500
    output_w = 500
    output_h = 500
    data_number = 5000
    file_name = dcgan.dataset_name
    file_type =dcgan.input_fname_pattern
    grayscale=False
    
    datafiles = glob(os.path.join("./data", file_name, file_type))
    inputs = dcgan.inputs
    
    DD_origin = []
    DD_logits = []
    DD_errD = []
    for idx in xrange(data_number):
      batch_files = datafiles[idx*batch_num:(idx+1)*batch_num]
      batch = [
          get_image(batch_file,
                    input_height=input_h,
                    input_width=input_w,
                    resize_height=output_h,
                    resize_width=output_w,
                    crop=True,
                    grayscale=False) for batch_file in batch_files]
      if grayscale:
        batch_images = np.array(batch).astype(np.float32)[:, :, :, None]
      else:
        batch_images = np.array(batch).astype(np.float32)
    
      errD_real = dcgan.d_loss_real.eval({ inputs: batch_images })
#      DD_errD[idx*batch_num:(idx+1)*batch_num] = errD_real
      
      D, D_logits = sess.run(dcgan.discriminator(inputs, reuse=True), feed_dict={inputs: batch_images})
      DD_origin[idx*batch_num:(idx+1)*batch_num] = D
      DD_logits[idx*batch_num:(idx+1)*batch_num] = D_logits
      
      print(" %d, %s, D_origin: %.8f, D_logits: %.8f, D_err: %.8f" % (idx, batch_files, D, D_logits, errD_real)) #need to change if batch_num > 1

    OutputFile = open('%s.txt' % (file_name), 'wb')
    np.savetxt(OutputFile, DD_logits, fmt='%.5f')
    
  elif option == 7:
      the_learning_rate = 0.1
      config.train_size = 100
      
      LOSS_R = []
      LOSS_G = []
      d_optim = tf.train.AdamOptimizer(the_learning_rate, beta1=config.beta1) \
                .minimize(dcgan.d_loss, var_list=dcgan.d_vars)
      g_optim = tf.train.AdamOptimizer(the_learning_rate, beta1=config.beta1) \
                .minimize(dcgan.g_loss, var_list=dcgan.g_vars)
      try:
        tf.global_variables_initializer().run()
      except:
        tf.initialize_all_variables().run()
        
      dcgan.g_sum = merge_summary([dcgan.z_sum, dcgan.d__sum,
      dcgan.G_sum, dcgan.d_loss_fake_sum, dcgan.g_loss_sum])
      dcgan.d_sum = merge_summary(
          [dcgan.z_sum, dcgan.d_sum, dcgan.d_loss_real_sum, dcgan.d_loss_sum])
    
      dcgan.data = glob(os.path.join(
        "./data", config.dataset, dcgan.input_fname_pattern))
      batch_idxs = min(len(dcgan.data), config.train_size) // config.batch_size
    
      for idx in xrange(0, batch_idxs):
        could_load, checkpoint_counter = dcgan.load(dcgan.checkpoint_dir)
        if could_load:
          counter = checkpoint_counter
          print(" [*] Load SUCCESS")
        else:
          print(" [!] Load failed...")
        
        batch_files = dcgan.data[idx*config.batch_size:(idx+1)*config.batch_size]
        batch = [
            get_image(batch_file,
                      input_height=dcgan.input_height,
                      input_width=dcgan.input_width,
                      resize_height=dcgan.output_height,
                      resize_width=dcgan.output_width,
                      crop=dcgan.crop,
                      grayscale=dcgan.grayscale) for batch_file in batch_files]
        if dcgan.grayscale:
          batch_images = np.array(batch).astype(np.float32)[:, :, :, None]
        else:
          batch_images = np.array(batch).astype(np.float32)
    
        batch_z = np.random.uniform(-1, 1, [config.batch_size, dcgan.z_dim]) \
              .astype(np.float32)
    
        # Update D network
        _, summary_str = dcgan.sess.run([d_optim, dcgan.d_sum],
          feed_dict={ dcgan.inputs: batch_images, dcgan.z: batch_z })
    #    dcgan.writer.add_summary(summary_str, counter)
    
        # Update G network
        _, summary_str = dcgan.sess.run([g_optim, dcgan.g_sum],
          feed_dict={ dcgan.z: batch_z })
    #    dcgan.writer.add_summary(summary_str, counter)
    
          
 #       errD_fake = dcgan.d_loss_fake.eval({ dcgan.z: batch_z })
        errD_real = dcgan.d_loss_real.eval({ dcgan.inputs: batch_images })
        errG = dcgan.g_loss.eval({dcgan.z: batch_z})
    
        print("file:%s, d_loss: %.8f, g_loss: %.8f" \
          % ( batch_files, errD_real, errG))
      
        LOSS_R[idx*config.batch_size:(idx+1)*config.batch_size] = [errD_real, errG]
#        LOSS_G[idx*config.batch_size:(idx+1)*config.batch_size] = [errG]

        samples, d_loss, g_loss = dcgan.sess.run(
          [dcgan.sampler, dcgan.d_loss, dcgan.g_loss],
          feed_dict={
              dcgan.z: batch_z,
              dcgan.inputs: batch_images,
          },
        )
        print("dadddadada:", samples.shape)
        save_images(samples, image_manifold_size(samples.shape[0]), './%s/test_arange_%s.png' % (config.sample_dir, idx))
          
      OutputFile = open('%Loss_s.txt' % (file_name), 'wb')
      np.savetxt(OutputFile, LOSS_R, fmt='%.5f')


  elif option == 8:
      for idx in xrange(0, 100):
        batch_z = np.random.uniform(-1, 1, [config.batch_size, dcgan.z_dim]) \
                .astype(np.float32)
        samples = sess.run(dcgan.sampler, feed_dict={dcgan.z: batch_z})
        print(batch_z)
        save_images(samples, image_manifold_size(samples.shape[0]), './%s/test_arange_%s.png' % (config.sample_dir, idx))

def image_manifold_size(num_images):
  manifold_h = int(np.floor(np.sqrt(num_images)))
  manifold_w = int(np.ceil(np.sqrt(num_images)))
  assert manifold_h * manifold_w == num_images
  return manifold_h, manifold_w


def sigmoid_cross_entropy_with_logits(x, y):
  try:
    return tf.nn.sigmoid_cross_entropy_with_logits(logits=x, labels=y)
  except:
    return tf.nn.sigmoid_cross_entropy_with_logits(logits=x, targets=y)
